/*
 * Copyright (C) 2006-2009 Andre Noll <maan@systemlinux.org>
 *
 * Licensed under the GPL v2. For licencing details see COPYING.
 */

/** \file dccp_send.c Paraslash's dccp sender. */

/*
 * based on server.c of dccp-cs-0.01.tar.bz2,
 * (C) 2005 Ian McDonald <imcdnzl@gmail.com>
 */

#include <regex.h>
#include <sys/types.h>
#include <dirent.h>
#include <osl.h>

#include "para.h"
#include "error.h"
#include "string.h"
#include "afh.h"
#include "afs.h"
#include "server.h"
#include "net.h"
#include "list.h"
#include "vss.h"
#include "send.h"
#include "fd.h"
#include "close_on_fork.h"
#include "chunk_queue.h"
#include "server.cmdline.h"
#include "acl.h"

/** Do not write more than that many bytes at once. */
#define DCCP_MAX_BYTES_PER_WRITE 1024

static struct sender_status dccp_sender_status, *dss = &dccp_sender_status;

static void dccp_pre_select(int *max_fileno, fd_set *rfds,
		__a_unused fd_set *wfds)
{
	if (dss->listen_fd >= 0)
		para_fd_set(dss->listen_fd, rfds, max_fileno);
}

static void dccp_post_select(fd_set *rfds, __a_unused fd_set *wfds)
{
	struct sender_client *sc;

	if (dss->listen_fd < 0 || !FD_ISSET(dss->listen_fd, rfds))
		return;
	sc = accept_sender_client(dss);
	if (!sc)
		return;
	/*
	 * Bypass unused CCID paths: the sender does not receive application data
	 * from the client; by shutting down this unused communication path we can
	 * reduce processing costs a bit. See analogous comment in dccp_recv.c.
	 */
	if (shutdown(sc->fd, SHUT_RD) >= 0)
		return;
	PARA_WARNING_LOG("%s\n", strerror(errno));
	shutdown_client(sc, dss);
}

static void dccp_send(long unsigned current_chunk,
		__a_unused long unsigned chunks_sent, const char *buf,
		size_t len, const char *header_buf, size_t header_len)
{
	struct sender_client *sc, *tmp;

	list_for_each_entry_safe(sc, tmp, &dss->client_list, node)
		send_chunk(sc, dss, DCCP_MAX_BYTES_PER_WRITE, current_chunk, buf,
			len, header_buf, header_len);
}

static void dccp_shutdown_clients(void)
{
	shutdown_clients(dss);
}

static int dccp_com_on(__a_unused struct sender_command_data *scd)
{
	return generic_com_on(dss, IPPROTO_DCCP);
}

static int dccp_com_off(__a_unused struct sender_command_data *scd)
{
	generic_com_off(dss);
	return 1;
}


static int dccp_com_deny(struct sender_command_data *scd)
{
	generic_com_deny(scd, dss);
	return 1;
}

static int dccp_com_allow(struct sender_command_data *scd)
{
	generic_com_allow(scd, dss);
	return 1;
}

static char *dccp_info(void)
{
	return get_sender_info(dss, "dccp");
}

/**
 * The init function of the dccp sender.
 *
 * \param s pointer to the dccp sender struct.
 *
 * It initializes all function pointers of \a s and starts
 * listening on the given port.
 */
void dccp_send_init(struct sender *s)
{
	int ret;

	s->info = dccp_info;
	s->send = dccp_send;
	s->pre_select = dccp_pre_select;
	s->post_select = dccp_post_select;
	s->shutdown_clients = dccp_shutdown_clients;
	s->help = generic_sender_help;
	s->client_cmds[SENDER_ON] = dccp_com_on;
	s->client_cmds[SENDER_OFF] = dccp_com_off;
	s->client_cmds[SENDER_DENY] = dccp_com_deny;
	s->client_cmds[SENDER_ALLOW] = dccp_com_allow;
	s->client_cmds[SENDER_ADD] = NULL;
	s->client_cmds[SENDER_DELETE] = NULL;

	init_sender_status(dss, conf.dccp_access_arg, conf.dccp_access_given,
		conf.dccp_port_arg, conf.dccp_max_clients_arg,
		conf.dccp_default_deny_given);
	ret = generic_com_on(dss, IPPROTO_DCCP);
	if (ret < 0)
		PARA_ERROR_LOG("%s\n", para_strerror(-ret));
}
